import tensorflow as tf


high_value = 1e16


def normalize_embeddings(x):
    return tf.linalg.l2_normalize(x, axis=1)


def make_adj_matrix(labels, dtype):
    lm = tf.cond(tf.equal(tf.rank(labels), 1), lambda: tf.expand_dims(labels, -1), lambda: labels)
    mem = tf.equal(lm, tf.transpose(lm))
    mem = tf.cast(mem, dtype)
    return mem


def make_adj_matrix1(labels, dtype):
    l, i = tf.unique(tf.squeeze(labels))
    onehot = tf.one_hot(i, depth=tf.size(l), dtype=dtype)
    mem = tf.matmul(onehot, tf.transpose(onehot))
    return mem


def cosine_dist(x):
    vn = normalize_embeddings(x)
    return 1. - tf.matmul(vn, tf.transpose(vn))


def euclidean_dist_normalized(x):
    cos = tf.maximum(cosine_dist(x), 1e-16)
    return tf.sqrt(2. * cos)


def euclidean_dist(x):
    x_l2 = tf.reduce_sum(x**2, axis=1, keepdims=True)
    d_sq = tf.math.add(x_l2, tf.transpose(x_l2)) - 2. * tf.matmul(x, tf.transpose(x))
    d_sq = tf.maximum(d_sq, 1e-16)
    return tf.sqrt(d_sq)


def hard_positive(x, adj):
    return tf.reduce_max(x * adj, axis=1)


def mask_negatives(x, adj):
    """
        returns a tensor of the same shape as x with all positive pair distances replaced with the
        biggest number in the tensor. This way, positive pairs will not be selected in reduce_min operation
    """
    # TODO: should not use this trick as it might introduce wrong gradients
    # choices = tf.reduce_sum(adj, axis=1)
    # # If below assertion fails, this means there was no negatives to select
    # assert_op = tf.Assert(tf.reduce_all(choices > 0.5), [adj])
    # return x * (1. - adj) + tf.reduce_max(x) * adj
    return x * (1. - adj) + high_value * adj

def hard_negative(x, adj):
    # tf.reduce_max(x) * x makes main diagonal very high, which is otherwize 0
    return tf.reduce_min(mask_negatives(x, adj), axis=1)


def mean_negative(x, adj, k=1):
    neg = mask_negatives(x, adj)
    neg, _ = tf.math.top_k(tf.negative(neg), k=k)
    neg = tf.negative(neg)
    mean_neg = tf.reduce_mean(neg, axis=-1)
    return mean_neg


def semihard_negative(x, adj, hard_pos, semi_margin=.0):
    hard_pos_2d = tf.expand_dims(hard_pos, axis=1)
    # mask_ij = True if a distance between i and j is longer than hard positive for i
    # other words all negatives that have mask = True are semi hard negatives and should be considered
    mask = x > hard_pos_2d + semi_margin
    # adj_ij = 1 for positive pairs and 0 for negative ones. We now need to pick those negatives that
    # have mask = 1
    mask = tf.cast((adj < 0.5) & mask, dtype=x.dtype)
    # TODO: in an unlucky event of really bad hard positive, there will be no hard negatives to select
    # TODO: How to deal with it? As of now, a high constant is selected (meaning that negative is very far
    # TODO: anf thus not making any gradients)
    return tf.reduce_min(mask_negatives(x, 1.-mask), axis=1)


def pairwise_distance(x, metric):
    if metric == 'euclidean':
        return euclidean_dist(x)
    elif metric == 'euclidean_norm':
        return euclidean_dist_normalized(x)
    elif metric == 'cosine':
        return cosine_dist(x)
    else:
        raise ValueError('Error: metric {} is not supported.'.format(metric))



import numpy as np
import tensorflow as tf
from triplet_tools import triplet_batch_priming_loss, triplet_batch_semihard_loss, triplet_batch_hard_loss


class TestSharedFunctions:

    emb = None
    labels = None

    def setup_method(self):
        self.emb = np.array([
            [.22, -.5, .11],
            [.3, -.66, .05],
            [.6, -0.27, .1],
            [.8, 0.2, .1],
            [.72, .24, .12]
        ])
        self.labels = np.array([1, 1, 1, 2, 2])

    def evaluate_tf_graph(self, g):
        if tuple([int(s) for s in tf.__version__.split('.')]) < (2, 0, 0):
            with tf.Session() as sess:
                value = g.eval()
        else:
            value = g
        return value

    def test_priming_loss_basic(self):
        triplet_loss = triplet_batch_priming_loss()
        g = triplet_loss(self.labels, self.emb)
        loss = self.evaluate_tf_graph(g)
        assert abs(loss - 0.30346566) < 0.00001

    def test_batch_hard_loss_basic(self):
        triplet_loss = triplet_batch_hard_loss(margin=0.3)
        g = triplet_loss(self.labels, self.emb)
        loss = self.evaluate_tf_graph(g)
        assert abs(loss - 0.07150213112) < 0.00001

    def test_batch_semihard_loss_basic(self):
        triplet_loss = triplet_batch_semihard_loss()
        g = triplet_loss(self.labels, self.emb)
        loss = self.evaluate_tf_graph(g)
        assert abs(loss - 0.5303652637399063) < 0.00001

    def test_batch_semihard_loss_with_semi_margin(self):
        triplet_loss = triplet_batch_semihard_loss(semi_margin=0.1)
        g = triplet_loss(self.labels, self.emb)
        loss = self.evaluate_tf_graph(g)
        assert abs(loss - 0.332636844872053) < 0.00001



import tensorflow as tf
import triplet_tools.losses.shared_functions as sf


def triplet_batch_hard_loss(margin=1.0, metric='euclidean_norm'):

    def loss_function(labels, embeddings):
        d = sf.pairwise_distance(embeddings, metric)
        adj = sf.make_adj_matrix(labels, embeddings.dtype)
        hp = sf.hard_positive(d, adj)
        hn = sf.hard_negative(d, adj)
        loss = tf.reduce_mean(tf.maximum(hp - hn + margin, 0.))
        return loss

    return loss_function



import tensorflow as tf
import importlib
import pytest
from triplet_tools import triplet_batch_semihard_loss, triplet_batch_priming_loss, triplet_batch_hard_loss


class TestLossFunctionsTensorflow:

    def setup_method(self):
        self.train_data, self.test_data = self.load_mnist()

    def load_mnist(self):
        mnist = tf.keras.datasets.mnist
        return mnist.load_data()

    def make_model(self):
        model = tf.keras.Sequential([
            tf.keras.layers.Flatten(input_shape=(28, 28)),
            tf.keras.layers.Dense(128, activation='relu'),
            tf.keras.layers.Dense(10, activation='linear')
        ])
        return model

    def test_batch_priming_loss(self):
        model = self.make_model()
        loss_func = triplet_batch_priming_loss()
        model.compile('adam', loss_func)
        hist = model.fit(self.train_data[0], self.train_data[1], epochs=2)
        loss_hist = hist.history['loss']
        assert loss_hist[-1] < 0.1
        assert loss_hist[0] > loss_hist[1]

    def test_batch_hard_loss(self):
        model = self.make_model()
        loss_func = triplet_batch_priming_loss()
        model.compile('adam', loss_func)
        hist = model.fit(self.train_data[0], self.train_data[1], epochs=1)
        bh_loss_func = triplet_batch_hard_loss()
        model.compile('adam', bh_loss_func)
        hist = model.fit(self.train_data[0], self.train_data[1], epochs=10, batch_size=100, shuffle=True)
        loss_hist = hist.history['loss']
        assert loss_hist[-1] < 0.5
        assert loss_hist[0] > loss_hist[1]



import numpy as np
import tensorflow as tf
import pytest
import triplet_tools.losses.shared_functions as sf


class TestSharedFunctions:

    emb = None
    labels = None

    def setup_method(self):
        self.emb = np.array([
            [.22, -.5, .11],
            [.3, -.66, .05],
            [.4, -.55, .2],
            [.8, 0.2, .1],
            [.72, .24, .12]
        ])
        self.labels = np.array([1, 1, 1, 2, 2])

    def evaluate_tf_graph(self, g):
        if tuple([int(s) for s in tf.__version__.split('.')]) < (2, 0, 0):
            with tf.Session() as sess:
                value = g.eval()
        else:
            value = g
        return value

    def test_normalization(self):
        g = sf.normalize_embeddings(self.emb)
        norm = self.evaluate_tf_graph(g)
        assert norm.shape == self.emb.shape
        assert np.equal(np.sum(norm ** 2, axis=1), np.array([1., 1., 1., 1., 1.])).all()

    def test_memebership(self):
        g = sf.make_adj_matrix(self.labels, self.emb.dtype)
        assert g.dtype == tf.float64
        mem = self.evaluate_tf_graph(g)
        assert mem.shape == (self.emb.shape[0], self.emb.shape[0])
        control = np.array([
            [1., 1., 1., 0., 0.],
            [1., 1., 1., 0., 0.],
            [1., 1., 1., 0., 0.],
            [0., 0., 0., 1., 1.],
            [0., 0., 0., 1., 1.],
        ])
        assert np.equal(mem, control).all()

    def test_cosine(self):
        g = sf.cosine_dist(self.emb)
        cos = self.evaluate_tf_graph(g)
        self.check_cosine_matrix(cos)

    def test_euclidean(self):
        g = sf.euclidean_dist(self.emb)
        ed = self.evaluate_tf_graph(g)
        self.check_euclidean_matrix(ed)

    def test_euclidean_norm(self):
        g = sf.euclidean_dist_normalized(self.emb)
        ed = self.evaluate_tf_graph(g)
        self.check_euclidean_norm_matrix(ed)

    def test_hard_pos(self):
        adj = sf.make_adj_matrix(self.labels, self.emb.dtype)
        dist = sf.euclidean_dist_normalized(self.emb)
        g = sf.hard_positive(dist, adj)
        hp = self.evaluate_tf_graph(g)
        assert hp.shape == (self.emb.shape[0],)
        control = np.array([2.25034931e-01, 2.93190947e-01, 2.93190947e-01, 8.41556606e-02, 8.41556606e-02])
        assert (np.abs(hp - control) < 1e-6).all()

    def test_mask_neg(self):
        adj = sf.make_adj_matrix(self.labels, self.emb.dtype)
        dist = sf.euclidean_dist_normalized(self.emb)
        g = sf.mask_negatives(dist, adj)
        mn = self.evaluate_tf_graph(g)
        assert mn.shape == (self.emb.shape[0], self.emb.shape[0])

    def test_hard_neg(self):
        adj = sf.make_adj_matrix(self.labels, self.emb.dtype)
        dist = sf.euclidean_dist_normalized(self.emb)
        g = sf.hard_negative(dist, adj)
        hn = self.evaluate_tf_graph(g)
        assert hn.shape == (self.emb.shape[0],)
        control = np.array([1.27439449e+00, 1.27499359e+00, 1.10399029e+00, 1.10399029e+00, 1.15717809e+00])
        assert (np.abs(hn - control) < 1e-6).all()

    def test_mean_neg(self):
        adj = sf.make_adj_matrix(self.labels, self.emb.dtype)
        dist = sf.euclidean_dist_normalized(self.emb)
        g = sf.mean_negative(dist, adj, 2)
        mn = self.evaluate_tf_graph(g)
        assert mn.shape == (self.emb.shape[0],)
        control = np.array([1.3003276, 1.3031166, 1.13058419, 1.18919239, 1.24171939])
        assert (np.abs(mn - control) < 1e-6).all()

    def test_semihard(self):
        emb = np.array([
            [.22, -.5, .11],
            [.3, -.66, .05],
            [.4, -.55, .2],
            [.2, -.5, .1],
            [.72, .24, .12]
        ])
        adj = sf.make_adj_matrix(self.labels, emb.dtype)
        dist = sf.euclidean_dist_normalized(emb)
        hp = sf.hard_positive(dist, adj)
        g = sf.semihard_negative(dist, adj, hp)
        shn = self.evaluate_tf_graph(g)
        assert shn.shape == (self.emb.shape[0],)
        control = np.array([1.32626070, 1.33123960, 1.15717809, sf.high_value, sf.high_value])
        assert (np.abs(shn - control) < 1e-6).all()

    def test_pairwise_distance(self):
        g = sf.pairwise_distance(self.emb, 'cosine')
        d = self.evaluate_tf_graph(g)
        self.check_cosine_matrix(d)
        g = sf.pairwise_distance(self.emb, 'euclidean')
        d = self.evaluate_tf_graph(g)
        self.check_euclidean_matrix(d)
        g = sf.pairwise_distance(self.emb, 'euclidean_norm')
        d = self.evaluate_tf_graph(g)
        self.check_euclidean_norm_matrix(d)
        with pytest.raises(ValueError):
            g = sf.pairwise_distance(self.emb, 'Non-supported')
            d = self.evaluate_tf_graph(g)

    def check_cosine_matrix(self, cos):
        assert cos.shape == (self.emb.shape[0], self.emb.shape[0])
        control = np.array([
             [0.00000000e+00, 8.49099331e-03, 2.53203600e-02, 8.12040663e-01, 8.79483718e-01],
             [8.49099331e-03, 1.11022302e-16, 4.29804658e-02, 8.12804323e-01, 8.86099440e-01],
             [2.53203600e-02, 4.29804658e-02, 1.11022302e-16, 6.09397285e-01, 6.69530565e-01],
             [8.12040663e-01, 8.12804323e-01, 6.09397285e-01, 0.00000000e+00, 3.54108761e-03],
             [8.79483718e-01, 8.86099440e-01, 6.69530565e-01, 3.54108761e-03, 0.00000000e+00]
        ])
        assert (np.abs(cos - control) < 1e-6).all()

    def check_euclidean_matrix(self, ed):
        assert ed.shape == (self.emb.shape[0], self.emb.shape[0])
        control = np.array([
            [0., 0.18867962, 0.20736441, 0.90912045, 0.89314053],
            [0.18867962, 0., 0.21118712, 0.99604217, 0.9956405],
            [0.20736441, 0.21118712, 0., 0.85586214, 0.85609579],
            [0.90912045, 0.99604217, 0.85586214, 0., 0.09165151],
            [0.89314053, 0.9956405, 0.85609579, 0.09165151, 0.]
        ])
        assert (np.abs(ed - control) < 1e-6).all()

    def check_euclidean_norm_matrix(self, ed):
        assert ed.shape == (self.emb.shape[0], self.emb.shape[0])
        control = np.array([
            [0.00000000e+00, 1.30314952e-01, 2.25034931e-01, 1.27439449e+00, 1.32626070e+00],
            [1.30314952e-01, 1.49011612e-08, 2.93190947e-01, 1.27499359e+00, 1.33123960e+00],
            [2.25034931e-01, 2.93190947e-01, 1.49011612e-08, 1.10399029e+00, 1.15717809e+00],
            [1.27439449e+00, 1.27499359e+00, 1.10399029e+00, 0.00000000e+00, 8.41556606e-02],
            [1.32626070e+00, 1.33123960e+00, 1.15717809e+00, 8.41556606e-02, 0.00000000e+00]
        ])
        assert (np.abs(ed - control) < 1e-6).all()



from .losses.triplet_batch_hard_loss import triplet_batch_hard_loss
from .losses.triplet_batch_semihard_loss import triplet_batch_semihard_loss
from .losses.triplet_batch_priming_loss import triplet_batch_priming_loss



from setuptools import setup
from setuptools import find_packages

long_description = '''
A set of tools for implementing triplet networks in tensorflow/keras. The toolbox includes a set 
of loss functions that plug in to tensorflow/keras neural network seamlessly, transforming
your model into a one-short learning triplet model  
'''

setup(name='triplet-tools',
      version='0.2.0',
      description='A toolbox for creating and training triplet networks in tensorflow',
      long_description=long_description,
      author='Maxim Scherbak',
      author_email='maxim.scherbak@gmail.com',
      url='https://github.com/maxsch3/triplet-toolbox',
      download_url='https://github.com/maxsch3/triplet-toolbox',
      license='MIT',
      install_requires=['tensorflow>=1.14.0'],
      extras_require={
          'tests': ['pytest',
                    'markdown'],
      },
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      packages=find_packages())


import tensorflow as tf
import triplet_tools.losses.shared_functions as sf


def triplet_batch_priming_loss(metric='euclidean_norm'):

    def loss_function(labels, embeddings):
        d = sf.pairwise_distance(embeddings, metric)
        adj = sf.make_adj_matrix(labels, embeddings.dtype)
        hp = sf.hard_positive(d, adj)
        hn = sf.hard_negative(d, adj)
        # pull hard positives to 0
        hp = hp**2
        # push hard negatives away from 0
        hn = tf.exp(10*tf.negative(hn))
        loss = tf.reduce_mean(hp + hn)
        return loss

    return loss_function
